////////////////////////////////////////////////////////////////////////////////
//     
//    __                                ______
//   / /_______________  ____  ___     / ____/
//  / __/ ___/ ___/ __ \/ __ \/ _ \   /___ )  
// / /_(__  ) /__/ /_/ / /_/ /  __/  ____/ /  
// \__/____/\___/\____/ .___/\___/  /_____/   
//                   /_/                      
//     
/// \file cedrusbox.c
/// Definitions of cedrusbox functions
/// \example cedrusbox01.c
/// \todo SV1 testen
////////////////////////////////////////////////////////////////////////////////


#include "../include/tscope5/cedrusbox.h"
#include "../include/tscope5/cedrusbox_internal.h"
#include "../include/tscope5/system_internal.h"


////////////////////////////////////////////////////////////////////////////////
/// @name Response registration
/// Cedrus response boxes can only be used as response devices. 
/// See timer.c for more information about response registration.
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Define a cedrusbox button as a response button.
///
/// \param device Number of the cedrusbox (devices are numbered from 1).
/// \param button Number of the button (buttons are counted from 1).
///
/// \return The reponse number associated with the button.
///
/// Give a positive number if you want to monitor button press events,
/// a negative number if you want to monitor button release events.
////////////////////////////////////////////////////////////////////////////////
int ts5_define_cedrusbox_button(int device, int button)
{
	ts5_check_cedrusbox("ts5_define_cedrusbox_button");
	ts5_log(TS5_LOGLEVEL_3, "ts5_define_cedrusbox_button(%d,%d)\n", device, button);

	if (device<1 || device>_ts5_status.timer.num_cedrusboxes) {
		ts5_fatal("%s: %s %d, %s %d\n", "ts5_define_cedrusbox_button",
					"device argument is", device, "number of cedrusboxes is", 
					_ts5_status.timer.num_cedrusboxes);
	}

	if (button==0) {
		ts5_fatal("%s: %s\n", "ts5_define_cedrusbox_button",
				 	"button argument is 0, response buttons are numbered from 1");
	}

	if (abs(button)>_ts5_status.timer.cedrusbox[device-1].num_buttons) {
		ts5_fatal("%s: %s %d, %s %d is %d\n", "ts5_define_cedrusbox_button",
					"button argument is ", button,
					"number of cedrusbox buttons for device", device, 
					_ts5_status.timer.cedrusbox[device-1].num_buttons);
	}

	if (button>0 && _ts5_status.timer.cedrusbox[device-1]
					.button_press_defined[button-1] !=0) {
		ts5_fatal("%s: %s %d %s\n", "ts5_define_cedrusbox_button",
					"button press", button, "is already defined");
	}

	if (button<0 && _ts5_status.timer.cedrusbox[device-1]
					.button_release_defined[-button-1]!=0) {
		ts5_fatal("%s: %s %d %s\n", "ts5_define_cedrusbox_button",
				 	"button release", button, "is already defined");
	}

	_ts5_status.timer.cedrusbox_is_response_device = 1;
	_ts5_status.timer.num_defined_buttons++;
	_ts5_status.timer.num_active_buttons++;
	_ts5_status.timer.cedrusbox[device-1].num_defined_buttons++;
	_ts5_status.timer.cedrusbox[device-1].num_active_buttons++;

	if (button>0) {
		_ts5_status.timer.cedrusbox[device-1]
			.button_press_defined[button-1] =
			_ts5_status.timer.num_defined_buttons;

		_ts5_status.timer.cedrusbox[device-1]
			.button_press_active[button-1] =
			_ts5_status.timer.num_defined_buttons;

	}
	else {
		_ts5_status.timer.cedrusbox[device-1]
			.button_release_defined[-button-1] =
			_ts5_status.timer.num_defined_buttons;

		_ts5_status.timer.cedrusbox[device-1]
			.button_release_active[-button-1] =
			_ts5_status.timer.num_defined_buttons;
	}

	return _ts5_status.timer.num_defined_buttons;
}


////////////////////////////////////////////////////////////////////////////////
/// Get then number of cedrusboxes that are connected to the system.
///
/// \return The number of cedrusboxes that are connected to the system.
////////////////////////////////////////////////////////////////////////////////
int ts5_get_num_cedrusboxes()
{
	ts5_check_cedrusbox("ts5_get_num_cedrusboxes");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_num_cedrusboxes()\n");
	
	return _ts5_status.timer.num_cedrusboxes;	
}


////////////////////////////////////////////////////////////////////////////////
/// Get then number of buttons available on a cedrusbox.
///
/// \param device Number of the cedrusbox (devices are numbered from 1).
///
/// \return The number of buttons on the cedrusbox.
////////////////////////////////////////////////////////////////////////////////
int ts5_get_num_cedrusbox_buttons(int device)
{
	ts5_check_cedrusbox("ts5_get_num_cedrusbox_buttons");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_num_cedrusbox_buttons(%d)\n", device);
	
	return _ts5_status.timer.cedrusbox[device-1].num_buttons;
}


////////////////////////////////////////////////////////////////////////////////
/// Set a parameter on a cedrusbox.
///
/// \param device Number of the cedrusbox.
/// \param parameter Code for the parameter that will be set.
/// \param value The value that will be set.
///
/// \return The previous value for the parameter.
///
/// The available commands for SV1 voice keys are:
/// - TS5_CEDRUSBOX_LOCKINGLEVEL: 
///   can be S5_CEDRUSBOX_LOCK or TS5_CEDRUSBOX_UNLOCK. 
///   This locks/unlocks the treshold and delay dials on the device.
/// - TS5_CEDRUSBOX_TRESHOLD: voice key treshold. 
///   Should be an integer between the values set with 
///   TS5_CEDRUSBOX_MIN_TRESHOLD and TS5_CEDRUSBOX_MAX_TRESHOLD.
/// - TS5_CEDRUSBOX_MIN_TRESHOLD: minimum voice key treshold that can be set.
///   Should be an integer value between 0 and the value set with
///   TS5_CEDRUSBOX_MAX_TRESHOLD.
/// - TS5_CEDRUSBOX_MAX_TRESHOLD: maximum voice key treshold that can be set.
///   Should be an integer value between the value set with
///   TS5_CEDRUSBOX_MIN_TRESHOLD and 255.
/// - TS5_CEDRUSBOX_RISE_DELAY: voice key rise delay in milliseconds.  
///   Should be an integer between the values set with
///   TS5_CEDRUSBOX_MIN_RISE_DELAY and TS5_CEDRUSBOX_MAX_RISE_DELAY.
/// - TS5_CEDRUSBOX_MIN_RISE_DELAY:	 minimum voice key rise delay 
///   that can be set. Should be an integer value between 0 and the value set
///   with TS5_CEDRUSBOX_MAX_RISE_DELAY.
/// - TS5_CEDRUSBOX_MAX_RISE_DELAY: maximum voice key rise delay 
///   that can be set. Should be an integer value between the value set 
///   with TS5_CEDRUSBOX_MIN_TRESHOLD and 100.
/// - TS5_CEDRUSBOX_DROP_DELAY: voice key drop delay in milliseconds. 
///   Should be an integer value between 0 and 255.
////////////////////////////////////////////////////////////////////////////////
int ts5_cedrusbox_set_parameter(int device, int parameter, int value)
{
	ts5_check_cedrusbox("ts5_cedrusbox_set_parameter");
	ts5_log(TS5_LOGLEVEL_4, "ts5_cedrusbox_set_parameter(%d,%d,%d)\n", 
				device, parameter, value);

	if (_ts5_status.timer.cedrusbox[device-1].type!=TS5_CEDRUSBOX_SV1) {
		ts5_fatal("%s: %s %d %s\n", "ts5_cedrusbox_set_parameter",
					"device", device, "is not a voice key");
	}

	// pause the response thread if necessary
	int pause_cedrusbox = _ts5_data.timer.cedrusbox_thread && 
							!_ts5_data.timer.cedrusbox_thread_is_paused;
	
	if (pause_cedrusbox) {
		al_join_thread(_ts5_data.timer.cedrusbox_thread, NULL);
		_ts5_data.timer.cedrusbox_thread_is_paused = 1;
	}

	int port_num = _ts5_status.timer.cedrusbox[device-1].port_num;
	int retval = ts5_cedrusbox_get_parameter(device, parameter);
	char command[6];

	int min, max;

	switch (parameter) {

	case TS5_CEDRUSBOX_LOCKINGLEVEL:

		if (value != TS5_CEDRUSBOX_LOCK && value != TS5_CEDRUSBOX_UNLOCK) {
			ts5_fatal("%s: %s %s\n", "ts5_cedrusbox_set_parameter",
					 	"TS5_CEDRUSBOX_LOCKINGLEVEL should be",
						"TS5_CEDRUSBOX_LOCK or TS5_CEDRUSBOX_UNLOCK");
		}

		sprintf(command, "f2%d", value);
		break;

	case TS5_CEDRUSBOX_TRESHOLD:

		min = ts5_cedrusbox_get_parameter(device, TS5_CEDRUSBOX_MIN_TRESHOLD);
		max = ts5_cedrusbox_get_parameter(device, TS5_CEDRUSBOX_MAX_TRESHOLD);

		if (value < min || value > max) {
			ts5_fatal("%s: %s min (%d) and max(%d)\n",  
						"ts5_cedrusbox_set_parameter",
						"treshold should be between", min, max);
		}

		sprintf(command, "b1%d", value);
		break;

	case TS5_CEDRUSBOX_MIN_TRESHOLD:

		max = ts5_cedrusbox_get_parameter(device, TS5_CEDRUSBOX_MAX_TRESHOLD);
			if (value < 0 || value > max) {
			ts5_fatal("%s: %s (%d)\n", "ts5_cedrusbox_set_parameter",
					 	"minimum treshold should be between 0 and max", max);
		}

		sprintf(command, "b6%d", value);
		break;

	case TS5_CEDRUSBOX_MAX_TRESHOLD:

		min = ts5_cedrusbox_get_parameter(device, TS5_CEDRUSBOX_MIN_TRESHOLD);

		if (value < min || value > 255) {
			ts5_fatal("%s: %s min (%d) and 255\n", 
						"ts5_cedrusbox_set_parameter",
					 	"maximum treshold should be between", min);
			}

			sprintf(command, "b7%d", value);
			break;

	case TS5_CEDRUSBOX_RISE_DELAY:

		min = ts5_cedrusbox_get_parameter(device, TS5_CEDRUSBOX_MIN_RISE_DELAY);
		max = ts5_cedrusbox_get_parameter(device, TS5_CEDRUSBOX_MAX_RISE_DELAY);

		if (value < min || value > max) {
			ts5_fatal("%s: %s min (%d) and max (%d)", 
						"ts5_cedrusbox_set_parameter",
						"rise delay should be between", min, max);
		}

		sprintf(command, "b2%d", value);
		break;

	case TS5_CEDRUSBOX_MIN_RISE_DELAY:

		max = ts5_cedrusbox_get_parameter(device, TS5_CEDRUSBOX_MAX_RISE_DELAY);

		if (value < 0 || value > max) {
			ts5_fatal("%s: %s (%d)\n", "ts5_cedrusbox_set_parameter",
					 	"minimum rise delay should be between 0 and max", max);
		}

		sprintf(command, "b4%d", value);
		break;

	case TS5_CEDRUSBOX_MAX_RISE_DELAY:

		min = ts5_cedrusbox_get_parameter(device, TS5_CEDRUSBOX_MIN_RISE_DELAY);

		if (value < min || value > 100) {
			ts5_fatal("%s: %s min (%d) and 100", 
						"ts5_cedrusbox_set_parameter",
						"maximum rise delay should be between", min);
		}

		sprintf(command, "b5%d", value);
		break;

	case TS5_CEDRUSBOX_DROP_DELAY:

		if (value < 0 || value > 255) {
			ts5_fatal("%s: %s\n", "ts5_cedrusbox_set_parameter",
							"drop delay should be between 0 and 255\n");
		}

		sprintf(command, "b3%d", value);
		break;

	default:
		ts5_fatal("ts5_cedrusbox_set_parameter: unknown parameter value (%d)\n", 
					parameter);
	}

	ts5_write_cedrusbox(port_num, command, 4);

	int newval = ts5_cedrusbox_get_parameter(device, parameter);

	if (newval != value) {
		ts5_fatal("%s: %s %d\n", "ts5_cedrusbox_set_parameter",
				 	"could not change parameter", parameter);
	}

	if (pause_cedrusbox) {
		al_start_thread(_ts5_data.timer.cedrusbox_thread);
		_ts5_data.timer.cedrusbox_thread_is_paused = 0;
	}

	return retval;
}


////////////////////////////////////////////////////////////////////////////////
/// Get a parameter on a cedrusbox.
///
/// \param device Number of the cedrusbox.
/// \param parameter Code for the parameter that will be queried.
///
/// \return The value for the parameter.
///
/// See ts5_cedrusbox_set_parameter() for the list of parameter values.
////////////////////////////////////////////////////////////////////////////////
int ts5_cedrusbox_get_parameter(int device, int parameter)
{
	ts5_check_cedrusbox("ts5_cedrusbox_get_parameter");
	ts5_log(TS5_LOGLEVEL_4, "ts5_cedrusbox_get_parameter(%d,%d)\n", 
				device, parameter);

	if (_ts5_status.timer.cedrusbox[device-1].type!=TS5_CEDRUSBOX_SV1) {
		ts5_fatal("ts5_cedrusbox_get_parameter: device %d is not a voice key\n", 
						device);
	}

	int port_num = _ts5_status.timer.cedrusbox[device-1].port_num;

	char command[4];

	switch (parameter) {

		case TS5_CEDRUSBOX_LOCKINGLEVEL:
			sprintf(command, "_f2");
			break;

		case TS5_CEDRUSBOX_TRESHOLD:
			sprintf(command, "_b1");
			break;

		case TS5_CEDRUSBOX_MIN_TRESHOLD:
			sprintf(command, "_b6");
			break;

		case TS5_CEDRUSBOX_MAX_TRESHOLD:
			sprintf(command, "_b7");
			break;

		case TS5_CEDRUSBOX_RISE_DELAY:
			sprintf(command, "_b2");
			break;

		case TS5_CEDRUSBOX_MIN_RISE_DELAY:
			sprintf(command, "_b4");
			break;

		case TS5_CEDRUSBOX_MAX_RISE_DELAY:
			sprintf(command, "_b5");
			break;

		case TS5_CEDRUSBOX_DROP_DELAY:
			sprintf(command, "_b3");
			break;

		default:
			ts5_fatal("%s: %s (%d)\n", "ts5_cedrusbox_get_parameter",
					 	"unknown parameter value", parameter);
	}

	// pause the response thread if necessary
	int pause_cedrusbox = _ts5_data.timer.cedrusbox_thread && 
						!_ts5_data.timer.cedrusbox_thread_is_paused;
	
	if (pause_cedrusbox) {
		al_join_thread(_ts5_data.timer.cedrusbox_thread, NULL);
		_ts5_data.timer.cedrusbox_thread_is_paused = 1;
	}
	
	// write command
	ts5_write_cedrusbox(port_num, command, 3);

	// wait for response
	char serial_buff[TS5_MAX_CHAR];
	serial_buff[0] = '\0';
	ts5_read_cedrusbox(port_num, serial_buff, 4);

	int value = 0;
	value = serial_buff[3];

	// restart thread if necessary
	if (pause_cedrusbox) {
		al_start_thread(_ts5_data.timer.cedrusbox_thread);
		_ts5_data.timer.cedrusbox_thread_is_paused = 0;
	}
	
	return value;
}


////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////
