////////////////////////////////////////////////////////////////////////////////
//     
//    __                                ______
//   / /_______________  ____  ___     / ____/
//  / __/ ___/ ___/ __ \/ __ \/ _ \   /___ )  
// / /_(__  ) /__/ /_/ / /_/ /  __/  ____/ /  
// \__/____/\___/\____/ .___/\___/  /_____/   
//                   /_/                      
//     
/// \file joystick.c
/// Definitions of joystick functions
/// \example joystick01.c
////////////////////////////////////////////////////////////////////////////////


#include "../include/tscope5/joystick.h"
#include "../include/tscope5/timer.h"
#include "../include/tscope5/joystick_internal.h"
#include "../include/tscope5/system_internal.h"


////////////////////////////////////////////////////////////////////////////////
/// @name Response registration
/// Joysticks can only be used as response devices. 
/// See timer.c for more information about response registration.
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Define a joystick button as a response button.
///
/// \param device Number of the joystick (devices are numbered from 1).
/// \param button Number of the button (buttons are counted from 1).
///
/// \return The reponse number associated with the button.
///
/// Give a positive number if you want to monitor button press events,
/// a negative number if you want to monitor button release events.
////////////////////////////////////////////////////////////////////////////////
int ts5_define_joystick_button(int device, int button)
{
	ts5_check_joystick("ts5_define_joystick_button");
	ts5_log(TS5_LOGLEVEL_3, "ts5_define_joystick_button(%d,%d)\n", device, button);

	if (device<1 || device>_ts5_status.timer.num_joysticks) {
		ts5_fatal("%s: %s %d, %s %d\n", "ts5_define_joystick_button",
				 	"device argument is", device,
					"number of joysticks is", 
					_ts5_status.timer.num_joysticks);
	}

	if (button==0) {
		ts5_fatal("%s: %s\n", "ts5_define_joystick_button",
				 	"button argument is 0, response buttons are numbered from 1");
	}

	if (abs(button)>_ts5_status.timer.joystick[device-1].num_buttons) {
		ts5_fatal("%s: %s %d, %s %d is %d\n", "ts5_define_joystick_button",
					"button argument is", button, 
					"number of joystick buttons for device", device,
					_ts5_status.timer.joystick[device-1].num_buttons);
	}

	if (button>0 && _ts5_status.timer.joystick[device-1]
					.button_press_defined[button-1]!=0) {
		ts5_fatal("%s: %s %d %s\n", "ts5_define_joystick_button",
				 	"button press", button, "is already defined");
	}

	if (button<0 && _ts5_status.timer.joystick[device-1]
					.button_release_defined[-button-1]!=0) {
		ts5_fatal("%s: %s %d %s\n", "ts5_define_joystick_button",
				 	"button release", button, "is already defined");
	}

	_ts5_status.timer.joystick_is_response_device = 1;
	_ts5_status.timer.num_defined_buttons++;
	_ts5_status.timer.num_active_buttons++;
	_ts5_status.timer.joystick[device-1].num_defined_buttons++;
	_ts5_status.timer.joystick[device-1].num_active_buttons++;

	if (button>0) {
		_ts5_status.timer.joystick[device-1]
			.button_press_defined[button-1] =
			_ts5_status.timer.num_defined_buttons;

		_ts5_status.timer.joystick[device-1]
			.button_press_active[button-1] =
			_ts5_status.timer.num_defined_buttons;
	}
	else {
		_ts5_status.timer.joystick[device-1]
			.button_release_defined[-button-1] =
			_ts5_status.timer.num_defined_buttons;

		_ts5_status.timer.joystick[device-1]
			.button_release_active[-button-1] =
			_ts5_status.timer.num_defined_buttons;
	}

	return _ts5_status.timer.num_defined_buttons;
}


////////////////////////////////////////////////////////////////////////////////
/// Get then number of joysticks that are connected to the system.
///
/// \return The number of joysticks that are connected to the system.
////////////////////////////////////////////////////////////////////////////////
int ts5_get_num_joysticks()
{
	ts5_check_joystick("ts5_get_num_joysticks");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_num_joysticks()\n");
	
	return _ts5_status.timer.num_joysticks;	
}


////////////////////////////////////////////////////////////////////////////////
/// Get then number of buttons available on a joystick.
///
/// \param device Number of the joystick (devices are numbered from 1).
///
/// \return The number of buttons on the joystick.
////////////////////////////////////////////////////////////////////////////////
int ts5_get_num_joystick_buttons(int device)
{
	ts5_check_joystick("ts5_get_num_joystick_buttons");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_num_joystick_buttons(%d)\n", device);
	
	return _ts5_status.timer.joystick[device-1].num_buttons;
}


////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////
