////////////////////////////////////////////////////////////////////////////////
//     
//    __                                ______
//   / /_______________  ____  ___     / ____/
//  / __/ ___/ ___/ __ \/ __ \/ _ \   /___ )  
// / /_(__  ) /__/ /_/ / /_/ /  __/  ____/ /  
// \__/____/\___/\____/ .___/\___/  /_____/   
//                   /_/                      
//     
/// \file joystick_internal.c
/// Definitions of internal joystick functions
////////////////////////////////////////////////////////////////////////////////

#include "../include/tscope5/joystick_internal.h"
#include "../include/tscope5/system_internal.h"
#include "../include/tscope5/timer_internal.h"

/// Is the joystick subsystem installed?
int  _ts5_is_joystick_installed = 0;

////////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each joystick function.
///
/// \param calling_function Name the function that calls for 
/// this check or installation.
///
/// Checks whether the joystick subsystem is installed. 
/// If not, the joystick subsystem is installed.
////////////////////////////////////////////////////////////////////////////////
void ts5_check_joystick(char *calling_function)
{
	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_joystick\n", calling_function);
	if (!_ts5_is_joystick_installed) {
		ts5_install_joystick(calling_function);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Install the joystick subsystem.
///
/// \param calling_function Name the function that calls for 
/// this check or installation.
///
/// This function is called automatically if necessary.
////////////////////////////////////////////////////////////////////////////////
void ts5_install_joystick(char *calling_function)
{
	if (!_ts5_is_timer_installed) {
		ts5_install_timer(calling_function);
	}

	if (!_ts5_is_joystick_installed) {

		_ts5_is_joystick_installed = 1;
		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5 joystick\n", 
				calling_function);

		if (!al_install_joystick()) {
			ts5_fatal("%s: could not install Tscope5 joystick\n", 
					calling_function);
		}

		al_register_event_source(_ts5_data.timer.response_queue,
									al_get_joystick_event_source());

		_ts5_status.timer.num_joysticks = al_get_num_joysticks();
		_ts5_status.timer.joystick =
			(TS5_JOYSTICK_STATUS *)al_malloc(sizeof(TS5_JOYSTICK_STATUS)*
			_ts5_status.timer.num_joysticks);

		int i;
		for (i=0; i<_ts5_status.timer.num_joysticks; i++) {

			ALLEGRO_JOYSTICK *joy = al_get_joystick(i);

			_ts5_status.timer.joystick[i].num_buttons = 
				al_get_joystick_num_buttons(joy);
			
			_ts5_status.timer.joystick[i].num_defined_buttons = 0;
			_ts5_status.timer.joystick[i].num_active_buttons = 0;

			_ts5_status.timer.joystick[i].button_press_defined =
				(int *)al_malloc(sizeof(int)
				* _ts5_status.timer.joystick[i].num_buttons);

			_ts5_status.timer.joystick[i].button_press_active =
				(int *)al_malloc(sizeof(int)
				* _ts5_status.timer.joystick[i].num_buttons);

			_ts5_status.timer.joystick[i].button_release_defined =
				(int *)al_malloc(sizeof(int)
				* _ts5_status.timer.joystick[i].num_buttons);

			_ts5_status.timer.joystick[i].button_release_active =
				(int *)al_malloc(sizeof(int)
				* _ts5_status.timer.joystick[i].num_buttons);

			int j;
			for (j=0; j<_ts5_status.timer.joystick[i].num_buttons; j++) {
				_ts5_status.timer.joystick[i].button_press_defined[j] = 0;
				_ts5_status.timer.joystick[i].button_press_active[j] = 0;
				_ts5_status.timer.joystick[i].button_release_defined[j] = 0;
				_ts5_status.timer.joystick[i].button_release_active[j] = 0;
			}
		}

		atexit(ts5_uninstall_joystick);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Uninstall the joystick subsystem.
///
/// This function is called automatically at the end of the program.
////////////////////////////////////////////////////////////////////////////////
void ts5_uninstall_joystick()
{
	if (_ts5_is_joystick_installed) {

		ts5_log(TS5_LOGLEVEL_1, "Uninstalling Tscope5 joystick\n");

		int i;
		for (i=0; i<_ts5_status.timer.num_joysticks; i++) {

			al_free(_ts5_status.timer.joystick[i].button_press_defined);
			_ts5_status.timer.joystick[i].button_press_defined = NULL;

			al_free(_ts5_status.timer.joystick[i].button_press_active);
			_ts5_status.timer.joystick[i].button_press_active = NULL;

			al_free(_ts5_status.timer.joystick[i].button_release_defined);
			_ts5_status.timer.joystick[i].button_release_defined = NULL;

			al_free(_ts5_status.timer.joystick[i].button_release_active);
			_ts5_status.timer.joystick[i].button_release_active = NULL;
		}

		al_free(_ts5_status.timer.joystick);
		_ts5_status.timer.joystick=NULL;

		_ts5_status.timer.num_joysticks = 0;
		_ts5_status.timer.joystick_is_response_device = 0;

		al_unregister_event_source(_ts5_data.timer.response_queue,
									al_get_joystick_event_source());

		#ifdef TS5_MACOSX
			al_uninstall_joystick();

		#endif

		_ts5_is_joystick_installed = 0;
	}
}

