////////////////////////////////////////////////////////////////////////////////
//     
//    __                                ______
//   / /_______________  ____  ___     / ____/
//  / __/ ___/ ___/ __ \/ __ \/ _ \   /___ )  
// / /_(__  ) /__/ /_/ / /_/ /  __/  ____/ /  
// \__/____/\___/\____/ .___/\___/  /_____/   
//                   /_/                      
//     
/// \file mouse_internal.c
/// Definitions of internal mouse functions
////////////////////////////////////////////////////////////////////////////////

#include "../include/tscope5/mouse_internal.h"
#include "../include/tscope5/system_internal.h"
#include "../include/tscope5/timer_internal.h"
#include "../include/tscope5/graphics_internal.h"

/// Is the mouse subsystem installed?
int  _ts5_is_mouse_installed = 0;


////////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each mouse function.
///
/// \param calling_function Name the function that calls for 
/// this check or installation.
///
/// Checks whether tscope5 is installed. If not, the program is aborted.
///
/// Checks whether a is display installed. If not, the program is aborted.
///
/// Checks whether the mouse subsystem installed. 
/// If not, the mouse subsystem is installed.
////////////////////////////////////////////////////////////////////////////////
void ts5_check_mouse(char *calling_function)
{
	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_mouse\n", calling_function);

	if (!_ts5_is_tscope5_installed) {
		ts5_fatal("%s: Tscope5 not installed\n", calling_function);
	}

	if (_ts5_status.active_display == -1) {
		ts5_fatal("%s: no active display available\n", calling_function);
	}

	if (!_ts5_is_mouse_installed) {
		ts5_install_mouse(calling_function);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Install the mouse subsystem.
///
/// \param calling_function Name the function that calls for 
/// this check or installation.
///
/// This function is called automatically if necessary.
////////////////////////////////////////////////////////////////////////////////
void ts5_install_mouse(char *calling_function)
{
	if (!_ts5_is_timer_installed) {
		ts5_install_timer(calling_function);
	}

	if (!_ts5_is_mouse_installed) {

		_ts5_is_mouse_installed = 1;
		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5 mouse\n", 
				calling_function);

		if (!al_install_mouse()) {
			ts5_fatal("%s: could not install Tscope5 mouse\n",
					calling_function);
		}

		al_register_event_source(_ts5_data.timer.response_queue,
									al_get_mouse_event_source());

		_ts5_status.timer.mouse.num_buttons = al_get_mouse_num_buttons();
		_ts5_status.timer.mouse.num_defined_buttons = 0;
		_ts5_status.timer.mouse.num_active_buttons = 0;

		_ts5_status.timer.mouse.button_press_defined =
			(int *)al_malloc(sizeof(int)*_ts5_status.timer.mouse.num_buttons);

		_ts5_status.timer.mouse.button_press_active =
			(int *)al_malloc(sizeof(int)*_ts5_status.timer.mouse.num_buttons);

		_ts5_status.timer.mouse.button_release_defined =
			(int *)al_malloc(sizeof(int)*_ts5_status.timer.mouse.num_buttons);

		_ts5_status.timer.mouse.button_release_active =
			(int *)al_malloc(sizeof(int)*_ts5_status.timer.mouse.num_buttons);

		int i;
		for (i=0; i<_ts5_status.timer.mouse.num_buttons; i++) {
			_ts5_status.timer.mouse.button_press_defined[i] = 0;
			_ts5_status.timer.mouse.button_press_active[i] = 0;
			_ts5_status.timer.mouse.button_release_defined[i] = 0;
			_ts5_status.timer.mouse.button_release_active[i] = 0;
		}

		atexit(ts5_uninstall_mouse);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Uninstall the mouse subsystem.
///
/// This function is called automatically at the end of the program.
////////////////////////////////////////////////////////////////////////////////
void ts5_uninstall_mouse()
{
	if (_ts5_is_mouse_installed) {

		ts5_log(TS5_LOGLEVEL_1, "Uninstalling Tscope5 mouse\n");

		_ts5_status.timer.mouse.num_buttons = 0;
		_ts5_status.timer.mouse.num_defined_buttons = 0;
		_ts5_status.timer.mouse.num_active_buttons = 0;

		al_free(_ts5_status.timer.mouse.button_press_defined);
		_ts5_status.timer.mouse.button_press_defined = NULL;

		al_free(_ts5_status.timer.mouse.button_press_active);
		_ts5_status.timer.mouse.button_press_active = NULL;

		al_free(_ts5_status.timer.mouse.button_release_defined);
		_ts5_status.timer.mouse.button_release_defined = NULL;

		al_free(_ts5_status.timer.mouse.button_release_active);
		_ts5_status.timer.mouse.button_release_active = NULL;

		_ts5_status.timer.mouse_is_response_device = 0;

		al_unregister_event_source(_ts5_data.timer.response_queue,
									al_get_mouse_event_source());

		al_uninstall_mouse();

		_ts5_is_mouse_installed = 0;
	}
}

