////////////////////////////////////////////////////////////////////////////////
//     
//    __                                ______
//   / /_______________  ____  ___     / ____/
//  / __/ ___/ ___/ __ \/ __ \/ _ \   /___ )  
// / /_(__  ) /__/ /_/ / /_/ /  __/  ____/ /  
// \__/____/\___/\____/ .___/\___/  /_____/   
//                   /_/                      
//     
/// \file primitives.c
/// Definitions of primitives drawing functions
/// \example primitives01.c
/// \example primitives02.c
/// \example primitives03.c
/// \example primitives04.c
////////////////////////////////////////////////////////////////////////////////


#include "../include/tscope5/primitives.h"
#include "../include/tscope5/graphics.h"
#include "../include/tscope5/primitives_internal.h"
#include "../include/tscope5/system_internal.h"
#include "../include/tscope5/graphics_internal.h"

#include <allegro5/allegro_primitives.h>


////////////////////////////////////////////////////////////////////////////////
/// @name Primitives drawing functions
/// The drawing functions all take the smallest possible number of parameters
/// to put something on the screen: some coordinates.
/// Drawing parameters are set with the graphics parameter functions.
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Draw a pixel.
///
/// \param x Horizontal position of the pixel.
/// \param y Vertical position of the pixel.
////////////////////////////////////////////////////////////////////////////////
void ts5_draw_pixel(double x, double y)
{
	ts5_check_primitives("ts5_draw_pixel");
	ts5_log(TS5_LOGLEVEL_5, "ts5_draw_pixel(%f,%f)\n", x, y);

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {
		x = ts5_relative_to_absolute_coordinate_x(x);
		y = ts5_relative_to_absolute_coordinate_y(y);
	}

	if (_ts5_status.graphics.coordinate_system 
			== TS5_CARTESIAN_COORDINATES) {
		x = ts5_cartesian_to_display_coordinate_x(x);
		y = ts5_cartesian_to_display_coordinate_y(y);
	}

	if (_ts5_status.graphics.drawing_thickness <= 0.0) {
		al_draw_pixel(x, y, _ts5_status.graphics.foreground_color);
	}
	else {
		al_draw_filled_circle(x, y, 
			_ts5_status.graphics.drawing_thickness,
			_ts5_status.graphics.foreground_color);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Get the color of a pixel.
///
/// \param x Horizontal position of the pixel.
/// \param y Vertical position of the pixel.
///
/// Return value:
/// TS5_COLOR structure.
////////////////////////////////////////////////////////////////////////////////
TS5_COLOR ts5_get_pixel_color(double x, double y)
{
	ts5_check_primitives("ts5_get_pixel_color");
	ts5_log(TS5_LOGLEVEL_5, "ts5_get_pixel_color(%f,%f)\n", x, y);

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {
		x = ts5_relative_to_absolute_coordinate_x(x);
		y = ts5_relative_to_absolute_coordinate_y(y);
	}

	if (_ts5_status.graphics.coordinate_system 
			== TS5_CARTESIAN_COORDINATES) {
		x = ts5_cartesian_to_display_coordinate_x(x);
		y = ts5_cartesian_to_display_coordinate_y(y);
	}

	return (al_get_pixel(_ts5_data.target, x, y));
}


////////////////////////////////////////////////////////////////////////////////
/// Draw a line segment.
///
/// \param x1 Horizontal position of the start point of the line segment.
/// \param y1 Vertical position of the start point of the line segment.
/// \param x2 Horizontal position of the end point of the line segment.
/// \param y2 Vertical position of the end point of the line segment.
////////////////////////////////////////////////////////////////////////////////
void ts5_draw_line(double x1, double y1, double x2, double y2)
{
	ts5_check_primitives("ts5_draw_line");
	ts5_log(TS5_LOGLEVEL_5, "ts5_draw_line(%f,%f,%f,%f)\n", x1, y1, x2, y2);

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {
		x1 = ts5_relative_to_absolute_coordinate_x(x1);
		x2 = ts5_relative_to_absolute_coordinate_x(x2);
		y1 = ts5_relative_to_absolute_coordinate_y(y1);
		y2 = ts5_relative_to_absolute_coordinate_y(y2);
	}

	if (_ts5_status.graphics.coordinate_system 
			== TS5_CARTESIAN_COORDINATES) {
		x1 = ts5_cartesian_to_display_coordinate_x(x1);
		x2 = ts5_cartesian_to_display_coordinate_x(x2);
		y1 = ts5_cartesian_to_display_coordinate_y(y1);
		y2 = ts5_cartesian_to_display_coordinate_y(y2);
	}

	al_draw_line(x1, y1, x2, y2, _ts5_status.graphics.foreground_color,
					_ts5_status.graphics.drawing_thickness);
}


////////////////////////////////////////////////////////////////////////////////
/// Draw a rectangle.
///
/// \param x1 Horizontal position of the first corner of the rectangle.
/// \param y1 Vertical position of the first corner of the rectangle.
/// \param x2 Horizontal position of the second corner of the rectangle.
/// \param y2 Vertical position of the second corner of the rectangle.
////////////////////////////////////////////////////////////////////////////////
void ts5_draw_rectangle(double x1, double y1, double x2, double y2)
{
	ts5_check_primitives("ts5_draw_rectangle");
	ts5_log(TS5_LOGLEVEL_5, "ts5_draw_rectangle(%f,%f,%f,%f)\n", 
				x1, y1, x2, y2);

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {
		x1 = ts5_relative_to_absolute_coordinate_x(x1);
		x2 = ts5_relative_to_absolute_coordinate_x(x2);
		y1 = ts5_relative_to_absolute_coordinate_y(y1);
		y2 = ts5_relative_to_absolute_coordinate_y(y2);
	}

	if (_ts5_status.graphics.coordinate_system 
			== TS5_CARTESIAN_COORDINATES) {
		x1 = ts5_cartesian_to_display_coordinate_x(x1);
		x2 = ts5_cartesian_to_display_coordinate_x(x2);
		y1 = ts5_cartesian_to_display_coordinate_y(y1);
		y2 = ts5_cartesian_to_display_coordinate_y(y2);
	}

	if (!_ts5_status.graphics.fill_mode) {
		al_draw_rectangle(x1, y1, x2, y2, 
			_ts5_status.graphics.foreground_color,
			_ts5_status.graphics.drawing_thickness);
	}
	else {
		al_draw_filled_rectangle(x1, y1, x2, y2, 
			_ts5_status.graphics.foreground_color);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Draw a rounded rectangle.
///
/// \param x1 Horizontal position of the first corner of the rectangle.
/// \param y1 Vertical position of the first corner of the rectangle.
/// \param x2 Horizontal position of the second corner of the rectangle.
/// \param y2 Vertical position of the second corner of the rectangle.
/// \param rx Horizontal radius of the round.
/// \param ry Vertical radius of the round.
////////////////////////////////////////////////////////////////////////////////
void ts5_draw_rounded_rectangle(double x1, double y1, double x2, double y2, 
									double rx, double ry)
{
	ts5_check_primitives("ts5_draw_rounded_rectangle");
	ts5_log(TS5_LOGLEVEL_5, "ts5_draw_rounded_rectangle(%f,%f,%f,%f,%f,%f)\n", 
				x1, y1, x2, y2, rx, ry);

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {
		x1 = ts5_relative_to_absolute_coordinate_x(x1);
		x2 = ts5_relative_to_absolute_coordinate_x(x2);
		y1 = ts5_relative_to_absolute_coordinate_y(y1);
		y2 = ts5_relative_to_absolute_coordinate_y(y2);
		rx = ts5_relative_to_absolute_coordinate_x(rx);
		ry = ts5_relative_to_absolute_coordinate_y(ry);
	}

	if (_ts5_status.graphics.coordinate_system 
			== TS5_CARTESIAN_COORDINATES) {
		x1 = ts5_cartesian_to_display_coordinate_x(x1);
		x2 = ts5_cartesian_to_display_coordinate_x(x2);
		y1 = ts5_cartesian_to_display_coordinate_y(y1);
		y2 = ts5_cartesian_to_display_coordinate_y(y2);
	}

	if (x1 > x2) {
		double t = x1;
		x1 = x2;
		x2 = t;
	}

	if (y1 > y2) {
		double t = y1;
		y1 = y2;
		y2 = t;
	}

	double w = x2 - x1;

	if (rx>w/2.0) {
		rx=w/2.0;
	}

	double h = y2 - y1;

	if (ry>h/2.0) {
		ry=h/2.0;
	}


	if (!_ts5_status.graphics.fill_mode) {
		al_draw_rounded_rectangle(x1, y1, x2, y2, rx, ry,
			_ts5_status.graphics.foreground_color,
			_ts5_status.graphics.drawing_thickness);
	}
	else {
		al_draw_filled_rounded_rectangle(x1, y1, x2, y2, rx, ry,
			_ts5_status.graphics.foreground_color);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Draw a triangle.
///
/// \param x1 Horizontal position of the first corner of the triangle.
/// \param y1 Vertical position of the first corner of the triangle.
/// \param x2 Horizontal position of the second corner of the triangle.
/// \param y2 Vertical position of the second corner of the triangle.
/// \param x3 Horizontal position of the third corner of the triangle.
/// \param y3 Vertical position of the third corner of the triangle.
////////////////////////////////////////////////////////////////////////////////
void ts5_draw_triangle(double x1, double y1, double x2, double y2, 
							double x3, double y3)
{
	ts5_check_primitives("ts5_draw_triangle");	
	ts5_log(TS5_LOGLEVEL_5, "ts5_draw_triangle(%f,%f,%f,%f,%f,%f)\n", 
				x1, y1, x2, y2, x3, y3);

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {
		x1 = ts5_relative_to_absolute_coordinate_x(x1);
		x2 = ts5_relative_to_absolute_coordinate_x(x2);
		x3 = ts5_relative_to_absolute_coordinate_x(x3);
		y1 = ts5_relative_to_absolute_coordinate_y(y1);
		y2 = ts5_relative_to_absolute_coordinate_y(y2);
		y3 = ts5_relative_to_absolute_coordinate_y(y3);
	}

	if (_ts5_status.graphics.coordinate_system 
			== TS5_CARTESIAN_COORDINATES) {
		x1 = ts5_cartesian_to_display_coordinate_x(x1);
		x2 = ts5_cartesian_to_display_coordinate_x(x2);
		x3 = ts5_cartesian_to_display_coordinate_x(x3);
		y1 = ts5_cartesian_to_display_coordinate_y(y1);
		y2 = ts5_cartesian_to_display_coordinate_y(y2);
		y3 = ts5_cartesian_to_display_coordinate_y(y3);
	}

	if (!_ts5_status.graphics.fill_mode) {
		al_draw_triangle(x1, y1, x2, y2, x3, y3,
							_ts5_status.graphics.foreground_color,
							_ts5_status.graphics.drawing_thickness);
	}
	else {
		al_draw_filled_triangle(x1, y1, x2, y2, x3, y3, 
			_ts5_status.graphics.foreground_color);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Draw a circle.
///
/// \param cx Horizontal position of the center of the circle.
/// \param cy Vertical position of the center of the circle.
/// \param r Radius of the circle.
////////////////////////////////////////////////////////////////////////////////
void ts5_draw_circle(double cx, double cy, double r)
{
	ts5_check_primitives("ts5_draw_circle");
	ts5_log(TS5_LOGLEVEL_5, "ts5_draw_circle(%f,%f,%f)\n", cx, cy, r);

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {
		cx = ts5_relative_to_absolute_coordinate_x(cx);
		cy = ts5_relative_to_absolute_coordinate_y(cy);
		r = ts5_relative_to_absolute_coordinate_x(r);
	}

	if (_ts5_status.graphics.coordinate_system 
			== TS5_CARTESIAN_COORDINATES) {
		cx = ts5_cartesian_to_display_coordinate_x(cx);
		cy = ts5_cartesian_to_display_coordinate_y(cy);
	}

	if (!_ts5_status.graphics.fill_mode) {
		al_draw_circle(cx, cy, r, _ts5_status.graphics.foreground_color,
						_ts5_status.graphics.drawing_thickness);
	}
	else {
		al_draw_filled_circle(cx, cy, r, 
			_ts5_status.graphics.foreground_color);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Draw an ellipse.
///
/// \param cx Horizontal position of the center of the ellipse.
/// \param cy Vertical position of the center of the ellipse.
/// \param rx Horizontal radius of the ellipse.
/// \param ry Vertical radius of the ellipse.
////////////////////////////////////////////////////////////////////////////////
void ts5_draw_ellipse(double cx, double cy, double rx, double ry)
{
	ts5_check_primitives("ts5_draw_ellipse");
	ts5_log(TS5_LOGLEVEL_5, "ts5_draw_ellipse(%f,%f,%f,%f)\n", cx, cy, rx, ry);

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {
		cx = ts5_relative_to_absolute_coordinate_x(cx);
		cy = ts5_relative_to_absolute_coordinate_y(cy);
		rx = ts5_relative_to_absolute_coordinate_x(rx);
		ry = ts5_relative_to_absolute_coordinate_y(ry);
	}

	if (_ts5_status.graphics.coordinate_system 
			== TS5_CARTESIAN_COORDINATES) {
		cx = ts5_cartesian_to_display_coordinate_x(cx);
		cy = ts5_cartesian_to_display_coordinate_y(cy);
	}

	if (!_ts5_status.graphics.fill_mode) {
		al_draw_ellipse(cx, cy, rx, ry, 
			_ts5_status.graphics.foreground_color,
			_ts5_status.graphics.drawing_thickness);
	}
	else {
		al_draw_filled_ellipse(cx, cy, rx, ry, 
			_ts5_status.graphics.foreground_color);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Draw an arc.
///
/// \param cx Horizontal position of the center of the arc.
/// \param cy Vertical position of the center of the arc.
/// \param r Radius of the arc.
/// \param start The initial angle of the arc (0.0 = west, 90.0 = north, ...).
/// \param delta Span of the arc (positive is counter clockwise, 
/// negative is clockwise).
////////////////////////////////////////////////////////////////////////////////
void ts5_draw_arc(double cx, double cy, double r, double start, double delta)
{
	ts5_check_primitives("ts5_draw_arc");
	ts5_log(TS5_LOGLEVEL_5, "ts5_draw_arc(%f,%f,%f,%f,%f)\n", 
				cx, cy, r, start, delta);

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {
		cx = ts5_relative_to_absolute_coordinate_x(cx);
		cy = ts5_relative_to_absolute_coordinate_y(cy);
		r = ts5_relative_to_absolute_coordinate_x(r);
	}

	if (_ts5_status.graphics.coordinate_system 
			== TS5_CARTESIAN_COORDINATES) {
		cx = ts5_cartesian_to_display_coordinate_x(cx);
		cy = ts5_cartesian_to_display_coordinate_y(cy);
	}

	start = - start * TS5_PI / 180;
	delta = - delta * TS5_PI / 180;
	al_draw_arc(cx, cy, r, start, delta, 
		_ts5_status.graphics.foreground_color,
		_ts5_status.graphics.drawing_thickness);
}


////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////
